import numpy as np
import torch
from numpy import linalg as npLA
from scipy import linalg as sciLA


class LA_Tools:

    def __init__(self, engine: str = "torch") -> None:
        self.engine = engine
        assert engine in [
            "torch",
            "numpy",
            "scipy",
        ], """
        Please select a valid computing engine... the options are:
        "torch", "numpy", "scipy"
        """

    @staticmethod
    def state_ordering(evecs_old, evecs_new, round=8, **kwargs):
        # Essentially a matrix of the fidelities: |<phi|psi>|
        overlap = np.round(abs(evecs_old @ evecs_new.T), round)
        # calculate trace distance
        # for o in overlap:
        #     for _o in o:
        #         if (_o>1):
        #             print('OVERLAP BIGGER THAN 1', _o)
        # dist = abs(1-overlap)/step**2
        # ordering = np.array([dist[i,:].argmin() for i in range(len(evecs_old))])  #python
        # ordering = np.argmin(dist,axis=1) #numpy
        ordering = np.argmax(overlap, axis=1)  # numpy
        return ordering

    @staticmethod
    def order_eig(evals, evecs, **kwargs):
        order = np.argsort(evals)
        evecs_ordered = evecs[order, :]
        evals_ordered = evals[order]
        return evals_ordered, evecs_ordered

    def diagonalize_batch(self, matrix_array, round=10, **kwargs):
        if self.engine == "numpy":
            w, v = npLA.eigh(matrix_array)
        elif self.engine == "scipy":
            w, v = sciLA.eigh(matrix_array)
        elif self.engine == "torch":
            w, v = torch.linalg.eigh(torch.from_numpy(matrix_array))
            w = w.numpy()
            v = v.numpy()
        evals_batch = np.round(w, round)
        evecs_batch = np.round(np.transpose(v, [0, 2, 1]), round)
        return evals_batch, evecs_batch

    def diagonalize(self, matrix, order=False, normalize=False, round=10, **kwargs):
        if self.engine == "numpy":
            w, v = npLA.eigh(matrix)
        elif self.engine == "scipy":
            w, v = sciLA.eigh(matrix)
        elif self.engine == "torch":
            w, v = torch.linalg.eigh(torch.from_numpy(matrix))
            w = w.numpy()
            v = v.numpy()
        evals = np.round(w, round)
        evecs = np.round(v.T, round)
        if normalize:
            for i, evec in enumerate(evecs):
                evecs[i] /= evec @ evec
        if order:
            idx_order = np.argsort(evals)
            evecs = evecs[idx_order, :]
            evals = evals[idx_order]
        return evals, evecs



from .energy_levels import (
    MoleculeLevels,
    branching_ratios,
    xa_branching_ratios,
    calculate_TDMs,
    calculate_TDM_evecs,
    calculate_forbidden_TDM_evecs,
    calculate_forbidden_TDMs,
)

from .hamiltonian_builders import (
    build_operator,
    tensor_matrix,
)

from . import matrix_elements_sym
from . import energy_levels
from . import hamiltonian_builders



from .src import (
    matrix_elements_sym,
    energy_levels,
    MoleculeLevels,
    hamiltonian_builders,
    branching_ratios,
    xa_branching_ratios,
    calculate_TDMs,
    calculate_TDM_evecs,
    calculate_forbidden_TDM_evecs,
    calculate_forbidden_TDMs,
    build_operator,
    tensor_matrix,
)

import numpy as np
import sympy as sy
import seaborn as sns
import latex
from IPython.display import Latex, display

sns.set()
sns.set_palette("terrain")
np.set_printoptions(precision=5, suppress=True)



import numpy as np


def count_momenta(J_list, M_values):
    if M_values == "all":
        num = 1
        for J in J_list:
            num *= 2 * J + 1
    else:
        J_coupled = recursive_add_J(J_list)
        if M_values == "pos":
            num = np.sum(J_coupled + 1 - (J_coupled[-1] % 1))
        elif M_values == "none":
            num = len(J_coupled)
    return int(num)


def recursive_add_J(J_list):
    n = len(J_list)
    if n == 2:
        J_sum = list_add_J(*J_list)
        return J_sum
    elif n > 2:
        J_main = J_list[:-1]
        J_add = J_list[-1]
        return list_add_J(recursive_add_J(J_main), J_add)


def base_add_J(J1, J2):
    J_sum = np.arange(abs(J1 - J2), J1 + J2 + 1, 1, dtype=np.float64)
    return J_sum


def list_add_J(J1, J2):  # J2 cannot be a list
    if isinstance(J1, (list, tuple, np.ndarray)) and not isinstance(
        J2, (list, tuple, np.ndarray)
    ):
        J_sum = []
        for _J in J1:
            J_sum.extend(base_add_J(_J, J2))
        return np.sort(np.array(J_sum))
    elif not isinstance(J1, (list, tuple, np.ndarray)) and not isinstance(
        J2, (list, tuple, np.ndarray)
    ):
        return base_add_J(J1, J2)
    elif isinstance(J2, (list, tuple, np.ndarray)):
        if not isinstance(J1, (list, tuple, np.ndarray)):
            return list_add_J(J2, J1)
        else:
            print("J1 and J2 cannot both be lists")
            return None


def q_numbers_bBJ_new(N_range, Lambda, S=1 / 2, I_list=[0, 1 / 2], M_values="all"):
    IYb = I_list[0]
    iH = I_list[-1]
    Nmin, Nmax = N_range[0], N_range[-1]
    if Nmin < abs(Lambda):
        print("Nmin must be >= L")
        Nmin = abs(Lambda)
    dim = 0
    for N in np.arange(Nmin, Nmax + 1, 1, dtype=np.float64):
        dim += count_momenta([N, S, *I_list], M_values)
    if Lambda != 0:
        dim *= 2
    if IYb == 0 or iH == 0:
        q_str = ["L", "N", "J", "F", "M"]
        I = max(IYb, iH)
        q_numbers = {}
        for q in q_str:
            q_numbers[q] = np.zeros(dim)
        i = 0
        for N in np.arange(Nmin, Nmax + 1, 1, dtype=np.float64):
            for J in np.arange(abs(N - S), abs(N + S) + 1, 1, dtype=np.float64):
                for F in np.arange(abs(J - I), abs(J + I) + 1, 1, dtype=np.float64):
                    if M_values == "none":
                        M = abs(F) % 1
                        for L in {True: [0], False: [-Lambda, Lambda]}[Lambda == 0]:
                            values = [L, N, J, F, M]
                            for q, val in zip(q_str, values):
                                q_numbers[q][i] = (
                                    val + 0
                                )  # looks weird but adding 0 converts -0 to 0
                            i += 1
                    else:
                        if M_values == "all":
                            Mmin = -F
                        elif M_values == "pos":
                            Mmin = abs(F) % 1
                        for M in np.arange(Mmin, F + 1, 1, dtype=np.float64):
                            for L in {True: [0], False: [-Lambda, Lambda]}[Lambda == 0]:
                                values = [L, N, J, F, M]
                                for q, val in zip(q_str, values):
                                    q_numbers[q][i] = (
                                        val + 0
                                    )  # looks weird but adding 0 converts -0 to 0
                                i += 1
    else:
        q_str = ["L", "N", "J", "F1", "F", "M"]
        q_numbers = {}
        for q in q_str:
            q_numbers[q] = np.zeros(dim)
        i = 0
        for N in np.arange(Nmin, Nmax + 1, 1, dtype=np.float64):
            for J in np.arange(abs(N - S), abs(N + S) + 1, 1, dtype=np.float64):
                for F1 in np.arange(
                    abs(J - IYb), abs(J + IYb) + 1, 1, dtype=np.float64
                ):
                    for F in np.arange(
                        abs(F1 - iH), abs(F1 + iH) + 1, 1, dtype=np.float64
                    ):
                        if M_values == "none":
                            M = abs(F) % 1
                            for L in {True: [0], False: [-Lambda, Lambda]}[Lambda == 0]:
                                values = [L, N, J, F1, F, M]
                                for q, val in zip(q_str, values):
                                    q_numbers[q][i] = (
                                        val + 0
                                    )  # looks weird but adding 0 converts -0 to 0
                                i += 1
                        else:
                            if M_values == "all":
                                Mmin = -F
                            elif M_values == "pos":
                                Mmin = abs(F) % 1
                            for M in np.arange(Mmin, F + 1, 1, dtype=np.float64):
                                for L in {True: [0], False: [-Lambda, Lambda]}[
                                    Lambda == 0
                                ]:
                                    values = [L, N, J, F1, F, M]
                                    for q, val in zip(q_str, values):
                                        q_numbers[q][i] = (
                                            val + 0
                                        )  # looks weird but adding 0 converts -0 to 0
                                    i += 1
    return q_numbers


def q_numbers_odd_bBJ(
    N_range, K_mag, S=1 / 2, I_list=[0, 1 / 2], M_values="all", M_range=[]
):
    IM = I_list[0]
    iH = I_list[-1]
    Nmin, Nmax = N_range[0], N_range[-1]
    K_mag = abs(K_mag)
    if Nmin < K_mag:
        print("Nmin must be >= |K|")
        Nmin = abs(K_mag)
    q_str = ["K", "N", "J", "F1", "F", "M"]
    q_numbers = {}
    for q in q_str:
        q_numbers[q] = []
    for N in np.arange(Nmin, Nmax + 1, 1, dtype=np.float64):
        for J in np.arange(abs(N - S), abs(N + S) + 1, 1, dtype=np.float64):
            for F1 in np.arange(abs(J - IM), abs(J + IM) + 1, 1, dtype=np.float64):
                for F in np.arange(abs(F1 - iH), abs(F1 + iH) + 1, 1, dtype=np.float64):
                    if M_values == "none":
                        for K in {True: [0], False: [-K_mag, K_mag]}[K_mag == 0]:
                            M = abs(F) % 1
                            values = [K, N, J, F1, F, M]
                            for q, val in zip(q_str, values):
                                q_numbers[q].append(
                                    val + 0
                                )  # looks weird but adding 0 converts -0 to 0
                    else:
                        if M_values == "all" or M_values == "custom":
                            Mmin = -F
                        elif M_values == "pos":
                            Mmin = abs(F) % 1
                        for M in np.arange(Mmin, F + 1, 1, dtype=np.float64):
                            if (
                                (M_values == "custom" and M in M_range)
                                or (M_values == "all")
                                or (M_values == "pos")
                            ):
                                for K in {True: [0], False: [-K_mag, K_mag]}[
                                    K_mag == 0
                                ]:
                                    values = [K, N, J, F1, F, M]
                                    for q, val in zip(q_str, values):
                                        q_numbers[q].append(
                                            val + 0
                                        )  # looks weird but adding 0 converts -0 to 0
                            elif M_values == "custom" and M not in M_range:
                                continue
    return q_numbers


def q_numbers_vibronic_even_bBJ(
    N_range,
    l_mag,
    L_mag,
    S=1 / 2,
    I_list=[0, 1 / 2],
    M_values="all",
    K_values=[],
    M_range=[],
):
    IM = I_list[0]
    iH = I_list[-1]
    Nmin, Nmax = N_range[0], N_range[-1]
    if K_values == []:
        K_values = [l + L for l in [-l_mag, l_mag] for L in [-L_mag, L_mag]]
    K_values = np.unique(K_values)
    K_min = abs(K_values).min()
    if Nmin < K_min:
        print("Nmin must be >= |Kmin|")
        Nmin = K_min
    q_str = ["l", "L", "K", "N", "J", "F", "M"]
    I = max(IM, iH)
    q_numbers = {}
    for q in q_str:
        q_numbers[q] = []
    for N in np.arange(Nmin, Nmax + 1, 1, dtype=np.float64):
        for J in np.arange(abs(N - S), abs(N + S) + 1, 1, dtype=np.float64):
            for F in np.arange(abs(J - I), abs(J + I) + 1, 1, dtype=np.float64):
                for K in K_values:
                    if N < abs(K):
                        continue
                    if M_values == "none":
                        M = abs(F) % 1
                        l_iter = {True: [0], False: [-l_mag, l_mag]}[l_mag == 0]
                        L_iter = {True: [0], False: [-L_mag, L_mag]}[L_mag == 0]
                        for l in l_iter:
                            for L in L_iter:
                                if K != l + L:
                                    continue
                                else:
                                    values = [l, L, K, N, J, F, M]
                                    for q, val in zip(q_str, values):
                                        q_numbers[q].append(
                                            val + 0
                                        )  # looks weird but adding 0 converts -0 to 0
                    else:
                        if M_values == "all" or M_values == "custom":
                            Mmin = -F
                        elif M_values == "pos":
                            Mmin = abs(F) % 1
                        for M in np.arange(Mmin, F + 1, 1, dtype=np.float64):
                            if (
                                (M_values == "custom" and M in M_range)
                                or (M_values == "all")
                                or (M_values == "pos")
                            ):
                                l_iter = {True: [0], False: [-l_mag, l_mag]}[l_mag == 0]
                                L_iter = {True: [0], False: [-L_mag, L_mag]}[L_mag == 0]
                                for l in l_iter:
                                    for L in L_iter:
                                        if K != l + L:
                                            continue
                                        else:
                                            values = [l, L, K, N, J, F, M]
                                            for q, val in zip(q_str, values):
                                                q_numbers[q].append(
                                                    val + 0
                                                )  # looks weird but adding 0 converts -0 to 0
                            elif M_values == "custom" and M not in M_range:
                                continue
    return q_numbers


def q_numbers_even_bBJ(
    N_range, K_mag, S=1 / 2, I_list=[0, 1 / 2], M_values="all", M_range=[]
):
    IM = I_list[0]
    iH = I_list[-1]
    Nmin, Nmax = N_range[0], N_range[-1]
    K_mag = abs(K_mag)
    if Nmin < K_mag:
        print("Nmin must be >= |K|")
        Nmin = abs(K_mag)
    q_str = ["K", "N", "J", "F", "M"]
    I = max(IM, iH)
    q_numbers = {}
    for q in q_str:
        q_numbers[q] = []
    for N in np.arange(Nmin, Nmax + 1, 1, dtype=np.float64):
        for J in np.arange(abs(N - S), abs(N + S) + 1, 1, dtype=np.float64):
            for F in np.arange(abs(J - I), abs(J + I) + 1, 1, dtype=np.float64):
                if M_values == "none":
                    for K in {True: [0], False: [-K_mag, K_mag]}[K_mag == 0]:
                        M = abs(F) % 1
                        values = [K, N, J, F, M]
                        for q, val in zip(q_str, values):
                            q_numbers[q].append(
                                val + 0
                            )  # looks weird but adding 0 converts -0 to 0
                else:
                    if M_values == "all" or M_values == "custom":
                        Mmin = -F
                    elif M_values == "pos":
                        Mmin = abs(F) % 1
                    for M in np.arange(Mmin, F + 1, 1, dtype=np.float64):
                        if (
                            (M_values == "custom" and M in M_range)
                            or (M_values == "all")
                            or (M_values == "pos")
                        ):
                            for K in {True: [0], False: [-K_mag, K_mag]}[K_mag == 0]:
                                values = [K, N, J, F, M]
                                for q, val in zip(q_str, values):
                                    q_numbers[q].append(
                                        val + 0
                                    )  # looks weird but adding 0 converts -0 to 0
                        elif M_values == "custom" and M not in M_range:
                            continue
    return q_numbers


def q_numbers_vibronic_even_aBJ(
    N_range,
    l_mag=0,
    L_mag=1,
    K_values=[],
    Omega_values=[],
    P_values=[1 / 2],
    S=1 / 2,
    I_list=[0, 1 / 2],
    M_values="all",
    M_range=[],
):  # Inputs need to be lists
    IM = I_list[0]
    iH = I_list[-1]
    Nmin, Nmax = N_range[0], N_range[-1]
    Jmin = abs(Nmin - S)
    Jmax = abs(Nmax + S)
    if not all(
        isinstance(values, list) for values in [Omega_values, P_values, K_values]
    ):
        print("Inputs for composite projection angular momenta must be lists")
        return
    l_iter = {True: [0], False: [-l_mag, l_mag]}[l_mag == 0]
    L_iter = {True: [0], False: [-L_mag, L_mag]}[L_mag == 0]
    S_iter = {True: [0], False: [-S, S]}[S == 0]
    if K_values == []:
        K_values = [l + L for l in l_iter for L in L_iter]
    if Omega_values == []:
        Omega_values = [_S + L for _S in S_iter for L in L_iter]
    if P_values == []:
        P_values = [l + L + _S for l in l_iter for L in L_iter for _S in S_iter]
    K_values = np.unique(np.abs(K_values))
    Omega_values = np.unique(np.abs(Omega_values))
    P_values = np.unique(np.abs(P_values))
    K_min = K_values.min()
    if Nmin < K_min:
        print("Nmin must be >= |Kmin|")
        Nmin = K_min
    q_str = ["l", "L", "K", "Sigma", "Omega", "P", "J", "F", "M"]
    q_numbers = {}
    for q in q_str:
        q_numbers[q] = []
    I = max(IM, iH)
    for J in np.arange(Jmin, Jmax + 1, 1, dtype=np.float64):
        for F in np.arange(abs(J - I), abs(J + I) + 1, 1, dtype=np.float64):
            for l in l_iter:
                for L in L_iter:
                    K = l + L
                    if abs(K) not in K_values:
                        continue
                    for Sigma in S_iter:
                        Omega = L + Sigma
                        if abs(Omega) not in Omega_values:
                            continue
                        P = K + Sigma
                        if abs(P) not in P_values:
                            continue
                        if J < abs(P):
                            continue
                        if M_values == "none":
                            M = abs(F) % 1
                            values = [l, L, K, Sigma, Omega, P, J, F, M]
                            for q, val in zip(q_str, values):
                                q_numbers[q].append(
                                    val + 0
                                )  # looks weird but adding 0 converts -0 to 0
                        else:
                            if M_values == "all" or M_values == "custom":
                                Mmin = -F
                            elif M_values == "pos":
                                Mmin = abs(F) % 1
                            for M in np.arange(Mmin, F + 1, 1, dtype=np.float64):
                                if (
                                    (M_values == "custom" and M in M_range)
                                    or (M_values == "all")
                                    or (M_values == "pos")
                                ):
                                    values = [l, L, K, Sigma, Omega, P, J, F, M]
                                    for q, val in zip(q_str, values):
                                        q_numbers[q].append(
                                            val + 0
                                        )  # looks weird but adding 0 converts -0 to 0
                                elif M_values == "custom" and M not in M_range:
                                    continue
    return q_numbers


def q_numbers_even_aBJ(
    N_range,
    K_mag=1,
    S=1 / 2,
    I_list=[0, 1 / 2],
    M_values="all",
    P_values=[1 / 2],
    M_range=[],
):
    IM = I_list[0]
    iH = I_list[-1]
    Nmin, Nmax = N_range[0], N_range[-1]
    Jmin = abs(Nmin - S)
    Jmax = abs(Nmax + S)
    K_mag = abs(K_mag)
    if Nmin < K_mag:
        print("Nmin must be >= |K|")
        Nmin = abs(K_mag)
    q_str = ["K", "Sigma", "P", "J", "F", "M"]
    q_numbers = {}
    for q in q_str:
        q_numbers[q] = []
    I = max(IM, iH)
    for J in np.arange(Jmin, Jmax + 1, 1, dtype=np.float64):
        for F in np.arange(abs(J - I), abs(J + I) + 1, 1, dtype=np.float64):
            if M_values == "none":
                for Sigma in np.arange(-abs(S), abs(S) + 1, 1, dtype=np.float64):
                    for K in {True: [0], False: [-K_mag, K_mag]}[K_mag == 0]:
                        P = K + Sigma
                        if abs(P) not in P_values:
                            continue
                        elif J < abs(P):
                            continue
                        else:
                            M = abs(F) % 1
                            values = [K, Sigma, P, J, F, M]
                        for q, val in zip(q_str, values):
                            q_numbers[q].append(
                                val + 0
                            )  # looks weird but adding 0 converts -0 to 0
            else:
                if M_values == "all" or M_values == "custom":
                    Mmin = -F
                elif M_values == "pos":
                    Mmin = abs(F) % 1
                for M in np.arange(Mmin, F + 1, 1, dtype=np.float64):
                    if (
                        (M_values == "custom" and M in M_range)
                        or (M_values == "all")
                        or (M_values == "pos")
                    ):
                        for Sigma in np.arange(
                            -abs(S), abs(S) + 1, 1, dtype=np.float64
                        ):
                            for K in {True: [0], False: [-K_mag, K_mag]}[K_mag == 0]:
                                P = K + Sigma
                                if abs(P) not in P_values:
                                    continue
                                elif J < abs(P):
                                    continue
                                else:
                                    values = [K, Sigma, P, J, F, M]
                                for q, val in zip(q_str, values):
                                    q_numbers[q].append(
                                        val + 0
                                    )  # looks weird but adding 0 converts -0 to 0
                    elif M_values == "custom" and M not in M_range:
                        continue
    return q_numbers


def q_numbers_odd_aBJ(
    N_range,
    K_mag=1,
    S=1 / 2,
    I_list=[5 / 2, 1 / 2],
    M_values="all",
    P_values=[1 / 2],
    M_range=[],
):
    IM = I_list[0]
    iH = I_list[-1]
    Nmin, Nmax = N_range[0], N_range[-1]
    K_mag = abs(K_mag)
    if Nmin < K_mag:
        print("Nmin must be >= |K|")
        Nmin = abs(K_mag)
    Jmin = abs(Nmin - S)
    Jmax = abs(Nmax + S)
    q_str = ["K", "Sigma", "P", "J", "F1", "F", "M"]
    q_numbers = {}
    for q in q_str:
        q_numbers[q] = []
    for J in np.arange(Jmin, Jmax + 1, 1, dtype=np.float64):
        for F1 in np.arange(abs(J - IM), abs(J + IM) + 1, 1, dtype=np.float64):
            for F in np.arange(abs(F1 - iH), abs(F1 + iH) + 1, 1, dtype=np.float64):
                if M_values == "none":
                    for Sigma in np.arange(-abs(S), abs(S) + 1, 1, dtype=np.float64):
                        for K in {True: [0], False: [-K_mag, K_mag]}[K_mag == 0]:
                            P = K + Sigma
                            if abs(P) not in P_values:
                                continue
                            else:
                                M = abs(F) % 1
                                values = [K, Sigma, P, J, F1, F, M]
                            for q, val in zip(q_str, values):
                                q_numbers[q].append(
                                    val + 0
                                )  # looks weird but adding 0 converts -0 to 0
                else:
                    if M_values == "all" or M_values == "custom":
                        Mmin = -F
                    elif M_values == "pos":
                        Mmin = abs(F) % 1
                    for M in np.arange(Mmin, F + 1, 1, dtype=np.float64):
                        if (
                            (M_values == "custom" and M in M_range)
                            or (M_values == "all")
                            or (M_values == "pos")
                        ):
                            for Sigma in np.arange(
                                -abs(S), abs(S) + 1, 1, dtype=np.float64
                            ):
                                for K in {True: [0], False: [-K_mag, K_mag]}[
                                    K_mag == 0
                                ]:
                                    P = K + Sigma
                                    if abs(P) not in P_values:
                                        continue
                                    else:
                                        values = [K, Sigma, P, J, F1, F, M]
                                    for q, val in zip(q_str, values):
                                        q_numbers[q].append(
                                            val + 0
                                        )  # looks weird but adding 0 converts -0 to 0
                        elif M_values == "custom" and M not in M_range:
                            continue
    return q_numbers


def q_numbers_decoupled(
    N_range, K_mag=0, S=1 / 2, I_list=[0, 1 / 2], M_values="all", M_range=[]
):
    IM = I_list[0]
    iH = I_list[-1]
    if M_values == "none":
        print("Cannot construct decoupled basis without M values")
        return {}
    Nmin, Nmax = N_range[0], N_range[-1]
    K_mag = abs(K_mag)
    if Nmin < K_mag:
        print("Nmin must be >= |K|")
        Nmin = abs(K_mag)
    if IM == 0 or iH == 0:
        I = max(IM, iH)
        q_str = ["K", "N", "M_N", "M_S", "M_I", "M_F"]
        q_numbers = {}
        for q in q_str:
            q_numbers[q] = []
        for N in np.arange(Nmin, Nmax + 1, 1, dtype=np.float64):
            for M_N in np.arange(-abs(N), abs(N) + 1, 1, dtype=np.float64):
                for M_S in np.arange(-abs(S), abs(S) + 1, 1, dtype=np.float64):
                    for M_I in np.arange(-abs(I), abs(I) + 1, 1, dtype=np.float64):
                        M_F = M_N + M_S + M_I
                        if (M_values == "pos" and M_F < 0) or (
                            M_values == "custom" and M_F not in M_range
                        ):
                            continue
                        else:
                            for K in {True: [0], False: [-K_mag, K_mag]}[K_mag == 0]:
                                values = [K, N, M_N, M_S, M_I, M_F]
                                for q, val in zip(q_str, values):
                                    q_numbers[q].append(
                                        val + 0
                                    )  # looks weird but adding 0 converts -0 to 0
    else:
        q_str = ["K", "N", "M_N", "M_S", "M_IM", "M_iH", "M_F"]
        q_numbers = {}
        for q in q_str:
            q_numbers[q] = []
        for N in np.arange(Nmin, Nmax + 1, 1, dtype=np.float64):
            for M_N in np.arange(-abs(N), abs(N) + 1, 1, dtype=np.float64):
                for M_S in np.arange(-abs(S), abs(S) + 1, 1, dtype=np.float64):
                    for M_IM in np.arange(-abs(IM), abs(IM) + 1, 1, dtype=np.float64):
                        for M_iH in np.arange(
                            -abs(iH), abs(iH) + 1, 1, dtype=np.float64
                        ):
                            M_F = M_N + M_S + M_IM + M_iH
                            if (M_values == "pos" and M_F < 0) or (
                                M_values == "custom" and M_F not in M_range
                            ):
                                continue
                            else:
                                for K in {True: [0], False: [-K_mag, K_mag]}[
                                    K_mag == 0
                                ]:
                                    values = [K, N, M_N, M_S, M_IM, M_iH, M_F]
                                    for q, val in zip(q_str, values):
                                        q_numbers[q].append(
                                            val + 0
                                        )  # looks weird but adding 0 converts -0 to 0
    return q_numbers


def q_numbers_decoupled_mJ(
    N_range, K_mag=0, S=1 / 2, I_list=[0, 1 / 2], M_values="all", M_range=[]
):
    IM = I_list[0]
    iH = I_list[-1]
    if M_values == "none":
        print("Cannot construct decoupled basis without M values")
        return {}
    Nmin, Nmax = N_range[0], N_range[-1]
    K_mag = abs(K_mag)
    if Nmin < K_mag:
        print("Nmin must be >= |K|")
        Nmin = abs(K_mag)
    # Jmin = abs(Nmin - S)
    # Jmax = abs(Nmax + S)
    if IM == 0 or iH == 0:
        I = max(IM, iH)
        q_str = ["K", "N", "J", "M_J", "M_I", "M_F"]
        q_numbers = {}
        for q in q_str:
            q_numbers[q] = []
        for N in np.arange(Nmin, Nmax + 1, 1, dtype=np.float64):
            for J in np.arange(abs(N - S), abs(N + S) + 1, 1, dtype=np.float64):
                for M_J in np.arange(-abs(J), abs(J) + 1, 1, dtype=np.float64):
                    for M_I in np.arange(-abs(I), abs(I) + 1, 1, dtype=np.float64):
                        M_F = M_J + M_I
                        if (M_values == "pos" and M_F < 0) or (
                            M_values == "custom" and M_F not in M_range
                        ):
                            continue
                        else:
                            for K in {True: [0], False: [-K_mag, K_mag]}[K_mag == 0]:
                                values = [K, N, J, M_J, M_I, M_F]
                                for q, val in zip(q_str, values):
                                    q_numbers[q].append(
                                        val + 0
                                    )  # looks weird but adding 0 converts -0 to 0
    else:
        q_str = ["K", "N", "J", "M_J", "M_IM", "M_iH", "M_F"]
        q_numbers = {}
        for q in q_str:
            q_numbers[q] = []
        for N in np.arange(Nmin, Nmax + 1, 1, dtype=np.float64):
            for J in np.arange(abs(N - S), abs(N + S) + 1, 1, dtype=np.float64):
                for M_J in np.arange(-abs(J), abs(J) + 1, 1, dtype=np.float64):
                    for M_IM in np.arange(-abs(IM), abs(IM) + 1, 1, dtype=np.float64):
                        for M_iH in np.arange(
                            -abs(iH), abs(iH) + 1, 1, dtype=np.float64
                        ):
                            M_F = M_J + M_IM + M_iH
                            if (M_values == "pos" and M_F < 0) or (
                                M_values == "custom" and M_F not in M_range
                            ):
                                continue
                            else:
                                for K in {True: [0], False: [-K_mag, K_mag]}[
                                    K_mag == 0
                                ]:
                                    values = [K, J, M_J, M_IM, M_iH, M_F]
                                    for q, val in zip(q_str, values):
                                        q_numbers[q].append(
                                            val + 0
                                        )  # looks weird but adding 0 converts -0 to 0
    return q_numbers


def q_numbers_bBS(
    N_range, K_mag, S=1 / 2, I_list=[5 / 2, 1 / 2], M_values="all", M_range=[]
):
    IM = I_list[0]
    iH = I_list[-1]
    if IM == 0 and iH != 0:
        IM = iH
        iH = 0
    Nmin, Nmax = N_range[0], N_range[-1]
    K_mag = abs(K_mag)
    if Nmin < K_mag:
        print("Nmin must be >= |K|")
        Nmin = abs(K_mag)
    # if IYb == 0 or iH == 0:
    #     I=max(IYb,iH)
    #     q_str = ['L','N','G','F','M']
    #     if M_values =='none':
    #         q_str = q_str[:-1]
    #     q_numbers = {}
    #     for q in q_str:
    #         q_numbers[q] = []
    #     for N in np.arange(Nmin,Nmax+1,1, dtype=np.float64):
    #         for G in np.arange(abs(I-S),abs(I+S)+1,1, dtype=np.float64):
    #             for F in np.arange(abs(G-N),abs(G+N)+1,1, dtype=np.float64):
    #                 if M_values=='none':
    #                     for L in {True:[0], False:[-Lambda,Lambda]}[Lambda==0]:
    #                         M=abs(F)%1
    #                         values = [L,N,G,F,M]
    #                         for q,val in zip(q_str,values):
    #                             q_numbers[q].append(val+0)
    #                 else:
    #                     if M_values=='all':
    #                         Mmin = -F
    #                     elif M_values=='pos':
    #                         Mmin = abs(F) % 1
    #                     for M in np.arange(Mmin,F+1,1, dtype=np.float64):
    #                         for L in {True:[0], False:[-Lambda,Lambda]}[Lambda==0]:
    #                             values = [L,N,G,F,M]
    #                             for q,val in zip(q_str,values):
    #                                 q_numbers[q].append(val+0)
    # else:
    q_str = ["K", "N", "G", "F1", "F", "M"]
    q_numbers = {}
    for q in q_str:
        q_numbers[q] = []
    for N in np.arange(Nmin, Nmax + 1, 1, dtype=np.float64):
        for G in np.arange(abs(IM - S), abs(IM + S) + 1, 1, dtype=np.float64):
            for F1 in np.arange(abs(G - N), abs(G + N) + 1, 1, dtype=np.float64):
                for F in np.arange(abs(F1 - iH), abs(F1 + iH) + 1, 1, dtype=np.float64):
                    if M_values == "none":
                        M = abs(F) % 1
                        for K in {True: [0], False: [-K_mag, K_mag]}[K_mag == 0]:
                            M = abs(F) % 1
                            values = [K, N, G, F1, F, M]
                            for q, val in zip(q_str, values):
                                q_numbers[q].append(
                                    val + 0
                                )  # looks weird but adding 0 converts -0 to 0
                    else:
                        if M_values == "all" or M_values == "custom":
                            Mmin = -F
                        elif M_values == "pos":
                            Mmin = abs(F) % 1
                        for M in np.arange(Mmin, F + 1, 1, dtype=np.float64):
                            if (
                                (M_values == "custom" and M in M_range)
                                or (M_values == "all")
                                or (M_values == "pos")
                            ):
                                for K in {True: [0], False: [-K_mag, K_mag]}[
                                    K_mag == 0
                                ]:
                                    values = [K, N, G, F1, F, M]
                                    for q, val in zip(q_str, values):
                                        q_numbers[q].append(
                                            val + 0
                                        )  # looks weird but adding 0 converts -0 to 0
                            elif M_values == "custom" and M not in M_range:
                                continue
    # else:
    #     I = max(I_list)
    #     q_str = ['K','N','G','F','M']
    #     q_numbers = {}
    #     for q in q_str:
    #         q_numbers[q] = []
    #     for N in np.arange(Nmin,Nmax+1,1, dtype=np.float64):
    #         for G in np.arange(abs(I-S),abs(I+S)+1,1, dtype=np.float64):
    #             for F in np.arange(abs(G-N),abs(G+N)+1,1, dtype=np.float64):
    #                 if M_values=='none':
    #                     M= abs(F) % 1
    #                     for K in {True:[0], False:[-K_mag,K_mag]}[K_mag==0]:
    #                         M=abs(F)%1
    #                         values = [K,N,G,F,M]
    #                         for q,val in zip(q_str,values):
    #                             q_numbers[q].append(val+0)
    #                 else:
    #                     if M_values=='all' or M_values=='custom':
    #                         Mmin = -F
    #                     elif M_values=='pos':
    #                         Mmin = abs(F) % 1
    #                     for M in np.arange(Mmin,F+1,1, dtype=np.float64):
    #                         if (
    # (M_values=='custom' and M in M_range) or (M_values=='all') or (M_values=='pos')
    # :
    #                             for K in {True:[0], False:[-K_mag,K_mag]}[K_mag==0]:
    #                                 values = [K,N,G,F,M]
    #                                 for q,val in zip(q_str,values):
    #                                     q_numbers[q].append(val+0)
    #                         elif M_values=='custom' and M not in M_range:
    #                             continue
    return q_numbers


#
# def q_numbers_173_decoupled(N_range,Lambda=0, S=1/2, I_list=[5/2,1/2],M_values='all'):
#     I = I_list[0]
#     iH = I_list[-1]
#     if M_values != 'all':
#         print('Cannot construct decoupled basis without M values')
#         return {}
#     else:
#         Nmin,Nmax=N_range[0],N_range[-1]
#         q_str = ['L','N','M_N','M_S','M_I','M_iH','M_F']
#         q_numbers = {}
#         for q in q_str:
#             q_numbers[q] = []
#         if Lambda==0:
#             for L in range(1, dtype=np.float64):
#                 for N in np.arange(Nmin,Nmax+1,1, dtype=np.float64):
#                     for M_N in np.arange(-abs(N),abs(N)+1,1, dtype=np.float64):
#                         for M_S in np.arange(-abs(S),abs(S)+1,1, dtype=np.float64):
#                             for M_I in np.arange(-abs(I),abs(I)+1,1, dtype=np.float64):
#                                 for M_iH in np.arange(-abs(iH),abs(iH)+1,1, dtype=np.float64):
#                                     M_F = M_N + M_S + M_I+ M_iH
#                                     if (M_values!='all' and M_F<0):
#                                         continue
#                                     else:
#                                         values = [L,N,M_N,M_S,M_I,M_iH,M_F]
#                                         for q,val in zip(q_str,values):
#                                             q_numbers[q].append(val+0)
#         elif Lambda==1:
#             if Nmin<abs(Lambda):
#                 print('Nmin must be >= L')
#                 Nmin=abs(Lambda)
#             for N in np.arange(Nmin,Nmax+1,1, dtype=np.float64):
#                 for M_N in np.arange(-abs(N),abs(N)+1,1, dtype=np.float64):
#                     for M_S in np.arange(-abs(S),abs(S)+1,1, dtype=np.float64):
#                         for M_I in np.arange(-abs(I),abs(I)+1,1, dtype=np.float64):
#                             for M_iH in np.arange(-abs(iH),abs(iH)+1,1, dtype=np.float64):
#                                 M_F = M_N + M_S + M_I+ M_iH
#                                 if (M_values!='all' and M_F<0):
#                                     continue
#                                 else:
#                                     for L in range(-1,1+1,2):
#                                         values = [L,N,M_N,M_S,M_I,M_iH,M_F]
#                                         for q,val in zip(q_str,values):
#                                             q_numbers[q].append(val+0)
#         return q_numbers



import molecular_structure.utils.matrix_elements as me
import molecular_structure.utils.quantum_numbers as qn
import molecular_structure.src.hamiltonian_builders as ham
from molecular_structure.config.molecule_parameters import all_params
from functools import partial


class Molecule_Library(object):
    """
    This class is primarily used as a library of functions and properties used
    to model the structure of YbOH. The class attributes are:

    matrix_elements: dictionary of relevant matrix elements, indexed by the
    YbOH state and isotope (ie, '174X000')

    cases: dictionary of the Hund's cases used for each state. The matrix
    elements are calculated in this basis.

    H_builders: dictionary of functions that construct the Hamiltonian matrix
    for a given isotope and state.

    PTV_builders: dictionary of functions that construct Hamiltonians for
    PT-violating effects.

    q_number_builders: dictionary of functions that construct the quantum
    numbers for a given state and isotope.
    The quantum numbers are represented as a dictionary, with the dict key
    giving the angular momentum name, and the dict value containing an array of
    the eigenvalues of each basis vector.
    For example, {'J': [0,1,1,1], 'M': [0,-1,0,1]}

    parameters: dictionary of the Hamiltonian constants associated with each
    state and isotope.

    Lambda: projection of angular momentum on the internuclear axis, excluding
    electron spin.
    """

    def __init__(self, molecule, I_spins, M_values, P_values, trap):
        self.I_spins = I_spins
        self.M_values = M_values
        self.P_values = P_values
        self.trap = trap
        self.molecule = molecule
        self.parameters = all_params[self.molecule]
        if len(P_values) == 1:
            self.parameters["ASO"] = 0
        self.matrix_elements = self.collect_all_matrix_elements(I_spins, M_values)
        self.cases = self.collect_all_cases()
        self.H_builders = self.collect_all_H_builders()
        self.PTV_builders = self.collect_all_PTV_builders()
        self.q_number_builders = self.collect_all_q_number_builders(I_spins, P_values)
        self.K = self.collect_all_K()
        self.basis_changers = self.collect_change_basis(I_spins)
        self.TDM_builders = self.collect_TDM()
        self.alt_q_number_builders = self.collect_alt_q(I_spins, P_values)
        self.TDM_p_builders = self.collect_p_TDM()
        self.all_parity = self.collect_parity()
        self.TDM_p_forbidden_builders = self.collect_p_TDM_forbidden()
        self.TDM_forbidden_builders = self.collect_TDM_forbidden()

    def collect_all_cases(self):
        if self.molecule == "YbOH":
            all_cases = {
                "174X000": "bBJ",
                "174X010": "bBJ",
                "173X000": "bBS",
                "173X010": "bBS",
                "171X000": "bBS",
                "171X010": "bBS",
                "174A000": "aBJ",
                "173A000": "aBJ",
                "171A000": "aBJ",
            }
        if self.molecule == "CaOH":
            all_cases = {
                "40X000": "bBJ",
                "40X010": "bBJ",
                "40A000": "aBJ",
                "40A010": "aBJ",
                "40B000": "bBJ",
            }
        return all_cases

    def collect_all_K(self):
        if self.molecule == "YbOH":
            all_K = {
                "174X000": 0,
                "174X010": 1,
                "173X000": 0,
                "171X000": 0,
                "173X010": 1,
                "171X010": 1,
                "174A000": 1,
                "173A000": 1,
                "171A000": 1,
            }
        if self.molecule == "CaOH":
            all_K = {
                "40X000": 0,
                "40X010": 1,
                "40A000": 1,
                "40A010": 2,
                "40B000": 0,
            }
        return all_K

    def collect_all_matrix_elements(self, I_spins, M_values):
        iH = self.I_spins[-1]
        IM = self.I_spins[0]

        bBJ_even_X_matrix_elements = {
            # Fine Structure
            "N^2": me.Rot_bBJ,  # N^2 Rotation
            "N.S": me.SR_bBJ,  # N.S Spin Rotation
            "l-doubling": me.lD_bBJ,  # Effective l doubling
            "l doubling p": me.p_lD_bBJ,
            "l doubling q": me.q_lD_bBJ,
            "NzSz": me.NzSz_bBJ,  # NzSz for bending mode
            # Hydrogen Hyperfine
            "I.S": me.IS_bBJ,  # I.S Fermi Contact Interaction
            "T2_0(I,S)": me.T2IS_bBJ,  # I S dipolar interaction
            "Iz": me.Iz_bBJ,  # I.n projection of I on internuclear axis n
            "Sz": me.Sz_bBJ,  # S.n projection of S on internuclear axis n
            "T2_2(I,S)": me.T2ISq2_bBJ,  # Anisotropic I.S dipolar
        }

        if M_values != "none":
            ext_fields = {
                # External Fields
                "ZeemanZ": me.ZeemanZ_bBJ,  # Zeeman interaction with lab z magnetic field
                "StarkZ": me.StarkZ_bBJ,  # Stark interaction with lab z electric field
                "ZeemanLZ": me.ZeemanLZ_bBJ,  # Including coupling to internuclear axis
                "ZeemanIZ": me.ZeemanIZ_bBJ,  # Nuclear spin coupling
            }
            bBJ_even_X_matrix_elements.update(ext_fields)
        if self.trap:
            trap_shifts = {
                "T0_azz": me.T00_alphaZZ_bBJ,
                "T0_axxyy": me.T00_alphaXXYY_bBJ,
                "T20_azz": partial(me.T20p_alphaZZ_bBJ, p=0),
                "T21_azz": partial(me.T20p_alphaZZ_bBJ, p=1),
                "T2-1_azz": partial(me.T20p_alphaZZ_bBJ, p=-1),
                "T22_azz": partial(me.T20p_alphaZZ_bBJ, p=2),
                "T2-2_azz": partial(me.T20p_alphaZZ_bBJ, p=-2),
                "T20_axxyy": partial(me.T20p_alphaXXYY_bBJ, p=0),
                "T21_axxyy": partial(me.T20p_alphaXXYY_bBJ, p=1),
                "T2-1_axxyy": partial(me.T20p_alphaXXYY_bBJ, p=-1),
                "T22_axxyy": partial(me.T20p_alphaXXYY_bBJ, p=2),
                "T2-2_axxyy": partial(me.T20p_alphaXXYY_bBJ, p=-2),
            }
            bBJ_even_X_matrix_elements.update(trap_shifts)
        for (
            term,
            element,
        ) in (
            bBJ_even_X_matrix_elements.items()
        ):  # iterate through, substitute hydrogen proton value
            bBJ_even_X_matrix_elements[term] = partial(element, I=iH)

        bBS_odd_X_matrix_elements = {
            # Fine Structure
            "N^2": me.Rot_bBS,  # N^2 Rotation
            "N.S": me.SR_bBS,  # N.S Spin Rotation
            "l-doubling": me.lD_bBS,  # Effective l doubling
            "NzSz": me.NzSz_bBS,  # NzSz for bending mode
            # Yb Hyperfine
            "IM.S": me.ISM_bBS,  # IM.S Fermi Contact Interaction
            "T2_0(IM,S)": me.T2ISM_bBS,  # q=0 molecule frame e-IM dd int (sqrt6 * (3IzSz - I.S))
            "T2_0(IM^2)": me.T2QM_bBS,  # q=0 molecule frame Yb e quad int (sqrt6 * (3Iz^2 - I^2))
            # Hydrogen Hyperfine
            "IH.S": me.ISH_bBS,  # IH.S Fermi Contact Interaction for Hydrogen
            "T2_0(IH,S)": me.T2ISH_bBS,  # q=0 molecule frame e-IH dd int (sqrt6 * (3IzSz - I.S))
        }

        if M_values != "none":
            ext_fields = {
                # External Fields
                "ZeemanZ": me.ZeemanZ_bBS,  # Zeeman interaction with lab z magnetic field
                "StarkZ": me.StarkZ_bBS,  # Stark interaction with lab z electric field
            }
            bBS_odd_X_matrix_elements.update(ext_fields)
        for (
            term,
            element,
        ) in bBS_odd_X_matrix_elements.items():  # substitue nuclear spin values
            bBS_odd_X_matrix_elements[term] = partial(element, iH=iH, I=IM)

        # bBS_odd_X_matrix_elements={
        # # Fine Structure
        # 'N^2': me.Rot_bBS,                 # N^2 Rotation
        # 'N.S': me.SR_bBS,                  # N.S Spin Rotation
        # 'l-doubling': me.lD_bBS,           # Effective l doubling
        # 'NzSz': me.NzSz_bBS,               # NzSz for bending mode
        #
        # # Yb Hyperfine
        # 'IM.S': me.ISM_bBS,              # IM.S Fermi Contact Interaction
        # 'T2_0(IM,S)': me.T2ISM_bBS,    # q=0 molecule frame e-IM dd int (sqrt6 * (3IzSz - I.S))
        # # 'T2_0(IM^2)': me.T2QM_bBS,   # q=0 molecule frame Yb e q int (sqrt6 * (3Iz^2 - I^2))
        #
        # # Hydrogen Hyperfine
        # 'IH.S': me.ISH_bBS,                # IH.S Fermi Contact Interaction for Hydrogen
        # 'T2_0(IH,S)': me.T2ISH_bBS   # q=0 molecule frame e-IH dd int (sqrt6 * (3IzSz - I.S))
        #
        # }
        #
        # if M_values != 'none':
        #     ext_fields = {
        #     # External Fields
        #     'ZeemanZ': me.ZeemanZ_bBS,         # Zeeman interaction with lab z magnetic field
        #     'StarkZ': me.StarkZ_bBS            # Stark interaction with lab z electric field
        #     }
        #     bBS_odd_X_matrix_elements.update(ext_fields)
        # for term, element in bBS_odd_X_matrix_elements.items():
        #      bBS_odd_X_matrix_elements[term] = partial(element,iH=iH,I=IM)

        aBJ_even_A_matrix_elements = {
            # Fine Structure
            "N^2": me.Rot_even_aBJ,  # N^2 Rotation
            "SO": me.SO_even_aBJ,  # Spin-Orbit
            "Lambda Doubling p+2q": me.LambdaDoubling_p2q_even_aBJ,  # Effective Lambda doubling
            "Lambda Doubling q": me.LambdaDoubling_q_even_aBJ,
            # Hydrogen Hyperfine
            "I.S": me.IS_even_aBJ,  # I.S Fermi Contact Interaction
            "IzSz": me.IzSz_even_aBJ,  # I.n*S.n projection of I and S on internuclear axis n
            "T2_0(IS)": me.T2IS_even_aBJ,  # q=0 molecule frame electron-IH (sqrt6 * (3IzSz - I.S))
        }

        if M_values != "none":
            ext_fields = {
                # External Fields
                "ZeemanLZ": me.ZeemanLZ_even_aBJ,
                "ZeemanSZ": me.ZeemanSZ_even_aBJ,
                "ZeemanParityZ": me.ZeemanParityZ_even_aBJ,
                "StarkZ": me.StarkZ_even_aBJ,  # Stark interaction with lab z electric field
            }
            aBJ_even_A_matrix_elements.update(ext_fields)

        for (
            term,
            element,
        ) in aBJ_even_A_matrix_elements.items():  # substitue nuclear spin values
            aBJ_even_A_matrix_elements[term] = partial(element, I=iH)

        aBJ_odd_A_matrix_elements = {
            # Fine Structure
            "N^2": me.Rot_odd_aBJ,  # N^2 Rotation
            "SO": me.SO_odd_aBJ,  # Spin Orbit
            "Lambda-Doubling": me.LambdaDoubling_odd_aBJ,  # Lambda doubling
            # Yb Hypeerfine
            "IzLz_M": me.ILM_odd_aBJ,
            "T2_2(IS)_M": me.T2q2_ISM_odd_aBJ,
            "T2_0(II)_M": me.T2q0_IM_odd_aBJ,
            # Hydrogen Hyperfine
            # 'IS_H': ,                   # I.S Fermi Contact Interaction
            # 'Iz_H': ,                   # I.n projection of I on internuclear axis n
            # 'Sz_H': ,                   # S.n projection of S on internuclear axis n
        }

        if M_values != "none":
            ext_fields = {
                # External Fields
                "ZeemanLZ": me.ZeemanLZ_odd_aBJ,
                "ZeemanSZ": me.ZeemanSZ_odd_aBJ,
                "ZeemanParityZ": me.ZeemanParityZ_odd_aBJ,
                "StarkZ": me.StarkZ_odd_aBJ,  # Stark interaction with lab z electric field
            }
            aBJ_odd_A_matrix_elements.update(ext_fields)
        for (
            term,
            element,
        ) in aBJ_odd_A_matrix_elements.items():  # substitue nuclear spin values
            aBJ_odd_A_matrix_elements[term] = partial(element, iH=iH, I=IM)

        # aBJ_odd_A_matrix_elements={
        # # Fine Structure
        # 'N^2': me.Rot_odd_aBJ,                 # N^2 Rotation
        # 'SO': me.SO_odd_aBJ,                    # Spin Orbit
        # 'Lambda-Doubling': me.LambdaDoubling_odd_aBJ,       #Lambda doubling
        #
        # # Yb Hypeerfine
        # 'IzLz_M': me.ILM_odd_aBJ,
        # 'T2_2(IS)_M': me.T2q2_ISM_odd_aBJ,
        # # 'T2_0(II)_M': me.T2q0_IM_odd_aBJ,
        #
        #
        # # Hydrogen Hyperfine
        # # 'IS_H': ,                   # I.S Fermi Contact Interaction
        # # 'Iz_H': ,                   # I.n projection of I on internuclear axis n
        # # 'Sz_H': ,                   # S.n projection of S on internuclear axis n
        # }
        #
        #
        # if M_values != 'none':
        #     ext_fields = {
        #     # External Fields
        #     'ZeemanLZ': me.ZeemanLZ_odd_aBJ,
        #     'ZeemanSZ': me.ZeemanSZ_odd_aBJ,
        #     'ZeemanParityZ': me.ZeemanParityZ_odd_aBJ,
        #     'StarkZ': me.StarkZ_odd_aBJ,            # Stark interaction with lab z electric field
        #     }
        #     aBJ_odd_A_matrix_elements.update(ext_fields)
        # for term, element in aBJ_odd_A_matrix_elements.items():
        #      aBJ_odd_A_matrix_elements[term] = partial(element,iH=iH,I=IM)

        if self.molecule == "YbOH":
            all_matrix_elements = {
                "174X000": bBJ_even_X_matrix_elements,
                "174X010": bBJ_even_X_matrix_elements,
                "173X000": bBS_odd_X_matrix_elements,
                "173X010": bBS_odd_X_matrix_elements,
                "174A000": aBJ_even_A_matrix_elements,
                "173A000": aBJ_odd_A_matrix_elements,
                "171A000": aBJ_odd_A_matrix_elements,
                "171X000": bBS_odd_X_matrix_elements,
                "171X010": bBS_odd_X_matrix_elements,
            }
        if self.molecule == "CaOH":
            all_matrix_elements = {
                "40X000": bBJ_even_X_matrix_elements,
                "40X010": bBJ_even_X_matrix_elements,
                "40A000": aBJ_even_A_matrix_elements,
                "40A010": aBJ_even_A_matrix_elements,
                "40B000": bBJ_even_X_matrix_elements,
            }

        return all_matrix_elements

    def collect_all_H_builders(self):
        if self.molecule == "YbOH":
            H_builders = {
                "174X000": ham.H_even_X,
                "174X010": ham.H_even_X,
                "173X000": ham.H_odd_X,
                "173X010": ham.H_odd_X,
                "174A000": ham.H_even_A,
                "173A000": ham.H_odd_A,
                "171A000": ham.H_odd_A,
                "171X000": ham.H_odd_X,
                "171X010": ham.H_odd_X,
            }
        if self.molecule == "CaOH":
            H_builders = {
                "40X000": ham.H_even_X,
                "40X010": partial(ham.H_even_X, trap=self.trap),
                "40A000": ham.H_even_A,
                "40B000": ham.H_even_X,
                # '40A010': ham.H_even_A
            }
        for key in H_builders:
            H_builders[key] = partial(
                H_builders[key], matrix_elements=self.matrix_elements[key]
            )
        return H_builders

    def collect_all_q_number_builders(self, I_spins, P_values):
        if self.molecule == "YbOH":
            q_number_builders = {
                "174X000": partial(qn.q_numbers_even_bBJ, K_mag=0),
                "174X010": partial(qn.q_numbers_even_bBJ, K_mag=1),
                "173X000": partial(qn.q_numbers_bBS, K_mag=0),
                "173X010": partial(qn.q_numbers_bBS, K_mag=1),
                "174A000": partial(qn.q_numbers_even_aBJ, K_mag=1, P_values=P_values),
                "173A000": partial(qn.q_numbers_odd_aBJ, K_mag=1, P_values=P_values),
                "171X000": partial(qn.q_numbers_bBS, K_mag=0),
                "171X010": partial(qn.q_numbers_bBS, K_mag=1),
                "171A000": partial(qn.q_numbers_odd_aBJ, K_mag=1, P_values=P_values),
                # '174aBJ': qn.q_numbers_174_aBJ,
                # '174bBJ': qn.q_numbers_bBJ,
            }
        if self.molecule == "CaOH":
            q_number_builders = {
                "40X000": partial(qn.q_numbers_even_bBJ, K_mag=0),
                "40X010": partial(qn.q_numbers_even_bBJ, K_mag=1),
                "40A000": partial(qn.q_numbers_even_aBJ, K_mag=1, P_values=P_values),
                "40A010": partial(qn.q_numbers_even_aBJ, K_mag=2, P_values=P_values),
                "40B000": partial(qn.q_numbers_even_bBJ, K_mag=0),
            }
        for key, builder in q_number_builders.items():
            q_number_builders[key] = partial(builder, I_list=self.I_spins)
        return q_number_builders

    def collect_all_PTV_builders(self):
        IM = self.I_spins[0]
        iH = self.I_spins[-1]

        if self.molecule == "YbOH":
            PTV_builders = {
                "174X000": ham.build_PTV_bBJ,
                "174X010": ham.build_PTV_bBJ,
                "173X000": ham.build_PTV_bBS,
                "173X010": partial(ham.build_PTV_bBS, IM=IM, iH=iH),
                "171X010": partial(ham.build_PTV_bBS, IM=IM, iH=iH),
            }
        if self.molecule == "CaOH":
            PTV_builders = {
                "40X000": ham.build_PTV_bBJ,
                "40X010": ham.build_PTV_bBJ,
            }
        return PTV_builders

    def collect_change_basis(self, I_spins):
        all_change_basis = {
            "a_bBJ": ham.convert_abBJ,
            "b_decoupled": partial(
                ham.decouple_b, I=max(I_spins)
            ),  # WiLL NOT WORK for case bBS
            "bBS_bBJ": partial(ham.convert_bbBS, I=max(I_spins)),
            "recouple_J": partial(ham.recouple_J, I=max(I_spins)),
            "b_I_decoupled": partial(ham.decouple_b_I, I=max(I_spins)),
            # 'A000_B010': partial(ham.build_operator,matrix_element = me.HRTSO_A000_B010),
            # 'A000_uA010': partial(ham.build_operator,matrix_element = me.HRTSO_A000_uA010),
            # 'A000_kA010': partial(ham.build_operator,matrix_element = me.HRTSO_A000_kA010),
            # 'X010_vibronic': partial(ham.build_operator,matrix_element =
            # me.convert_X010_vibronic),
            # 'A000_vibronic': partial(ham.build_operator,matrix_element =
            # me.convert_A000_vibronic)
        }
        return all_change_basis

    def collect_p_TDM(self):
        IM = self.I_spins[0]
        iH = self.I_spins[-1]
        if self.M_values != "none":
            if self.molecule == "YbOH":
                p_TDM = {
                    "174A000": partial(
                        ham.build_p_TDM_aBJ,
                        TDM_matrix_element=partial(me.TDM_p_even_aBJ, I=iH),
                    ),
                    "174X010": partial(
                        ham.build_p_TDM_aBJ,
                        TDM_matrix_element=partial(me.TDM_p_even_aBJ, I=iH),
                    ),
                    "173A000": partial(
                        ham.build_p_TDM_aBJ,
                        TDM_matrix_element=partial(me.TDM_p_odd_aBJ, iH=iH, I=IM),
                    ),
                    "171A000": partial(
                        ham.build_p_TDM_aBJ,
                        TDM_matrix_element=partial(me.TDM_p_odd_aBJ, iH=iH, I=IM),
                    ),
                }
            if self.molecule == "CaOH":
                p_TDM = {
                    "40A000": partial(
                        ham.build_p_TDM_aBJ,
                        TDM_matrix_element=partial(me.TDM_p_even_aBJ, I=iH),
                    ),
                    "40B000": partial(
                        ham.build_p_TDM_aBJ,
                        TDM_matrix_element=partial(me.TDM_p_even_aBJ, I=iH),
                    ),
                    "40X010": partial(
                        ham.build_p_TDM_aBJ,
                        TDM_matrix_element=partial(me.TDM_p_even_aBJ, I=iH),
                    ),
                    "40X000": partial(
                        ham.build_p_TDM_aBJ,
                        TDM_matrix_element=partial(me.TDM_p_even_aBJ, I=iH),
                    ),
                }
            return p_TDM
        else:
            return None

    def collect_p_TDM_forbidden(self):
        # IM = self.I_spins[0]
        iH = self.I_spins[-1]
        if self.M_values != "none":
            if self.molecule == "YbOH":
                p_TDM = {
                    "174A000": partial(
                        ham.build_TDM_aBJ_forbidden,
                        TDM_matrix_element=partial(me.TDM_p_vibronic_aBJ, I=iH),
                    ),
                    "uA010": partial(
                        ham.build_TDM_aBJ_forbidden,
                        TDM_matrix_element=partial(me.TDM_p_uA010_aBJ, I=iH),
                    ),
                    "kA010": partial(
                        ham.build_TDM_aBJ_forbidden,
                        TDM_matrix_element=partial(me.TDM_p_kA010_aBJ, I=iH),
                    ),
                    "B010": partial(
                        ham.build_TDM_aBJ_forbidden,
                        TDM_matrix_element=partial(me.TDM_p_B010_aBJ, I=iH),
                    ),
                }
            if self.molecule == "CaOH":
                p_TDM = {}
            return p_TDM
        else:
            return None

    def collect_TDM(self):
        IM = self.I_spins[0]
        iH = self.I_spins[-1]
        if self.M_values != "none":
            if self.molecule == "YbOH":
                all_TDM = {
                    "174A000": partial(
                        ham.build_TDM_aBJ,
                        TDM_matrix_element=partial(me.TransitionDipole_even_aBJ, I=iH),
                    ),
                    "173A000": partial(
                        ham.build_TDM_aBJ,
                        TDM_matrix_element=partial(
                            me.TransitionDipole_odd_aBJ, iH=iH, I=IM
                        ),
                    ),
                    "171A000": partial(
                        ham.build_TDM_aBJ,
                        TDM_matrix_element=partial(
                            me.TransitionDipole_odd_aBJ, iH=iH, I=IM
                        ),
                    ),
                }
            if self.molecule == "CaOH":
                all_TDM = {
                    "40A000": partial(
                        ham.build_TDM_aBJ,
                        TDM_matrix_element=partial(me.TransitionDipole_even_aBJ, I=iH),
                    ),
                    "40B000": partial(
                        ham.build_TDM_aBJ,
                        TDM_matrix_element=partial(me.TransitionDipole_even_aBJ, I=iH),
                    ),
                }
        else:
            if self.molecule == "YbOH":
                all_TDM = {
                    "174A000": partial(
                        ham.build_TDM_aBJ,
                        TDM_matrix_element=partial(
                            me.TransitionDipole_even_aBJ_noM, I=iH
                        ),
                    ),
                    "173A000": partial(
                        ham.build_TDM_aBJ,
                        TDM_matrix_element=partial(
                            me.TransitionDipole_odd_aBJ_noM, iH=iH, I=IM
                        ),
                    ),
                    "171A000": partial(
                        ham.build_TDM_aBJ,
                        TDM_matrix_element=partial(
                            me.TransitionDipole_odd_aBJ_noM, iH=iH, I=IM
                        ),
                    ),
                }
            if self.molecule == "CaOH":
                all_TDM = {
                    "40A000": partial(
                        ham.build_TDM_aBJ,
                        TDM_matrix_element=partial(
                            me.TransitionDipole_even_aBJ_noM, I=iH
                        ),
                    ),
                    "40B000": partial(
                        ham.build_TDM_aBJ,
                        TDM_matrix_element=partial(
                            me.TransitionDipole_even_aBJ_noM, I=iH
                        ),
                    ),
                }
        return all_TDM

    def collect_TDM_forbidden(self):
        # IM = self.I_spins[0]
        iH = self.I_spins[-1]
        if self.M_values != "none":
            if self.molecule == "YbOH":
                all_TDM = {
                    "174A000": partial(
                        ham.build_operator,
                        matrix_element=partial(
                            me.TransitionDipole_even_aBJ_vibronic, I=iH
                        ),
                    ),
                }
            if self.molecule == "CaOH":
                all_TDM = {}
        else:
            if self.molecule == "YbOH":
                all_TDM = {
                    "174A000": partial(
                        ham.build_TDM_aBJ_forbidden,
                        TDM_matrix_element=partial(
                            me.TransitionDipole_aBJ_vibronic_noM, I=iH
                        ),
                    ),
                }
            if self.molecule == "CaOH":
                all_TDM = {}
        return all_TDM

    def collect_TDM_2(self):
        iH = self.I_spins[-1]
        IM = self.I_spins[0]
        if self.M_values != "none":
            if self.molecule == "YbOH":
                all_TDM = {
                    "174A000": partial(
                        ham.build_TDM_aBJ,
                        TDM_matrix_element=partial(me.TransitionDipole_even_aBJ, I=iH),
                    ),
                    "173A000": partial(
                        ham.build_TDM_aBJ,
                        TDM_matrix_element=partial(
                            me.TransitionDipole_odd_aBJ, iH=iH, I=IM
                        ),
                    ),
                    "171A000": partial(
                        ham.build_TDM_aBJ,
                        TDM_matrix_element=partial(
                            me.TransitionDipole_odd_aBJ, iH=iH, I=IM
                        ),
                    ),
                }
            if self.molecule == "CaOH":
                all_TDM = {
                    "40A000": partial(
                        ham.build_TDM_aBJ,
                        TDM_matrix_element=partial(me.TransitionDipole_even_aBJ, I=iH),
                    ),
                    "40B000": partial(
                        ham.build_TDM_aBJ,
                        TDM_matrix_element=partial(me.TransitionDipole_even_aBJ, I=iH),
                    ),
                }
        else:
            if self.molecule == "YbOH":
                all_TDM = {
                    "174A000": partial(
                        ham.build_TDM_aBJ,
                        TDM_matrix_element=partial(
                            me.TransitionDipole_even_aBJ_noM, I=iH
                        ),
                    ),
                    "173A000": partial(
                        ham.build_TDM_aBJ,
                        TDM_matrix_element=partial(
                            me.TransitionDipole_odd_aBJ_noM, iH=iH, I=IM
                        ),
                    ),
                    "171A000": partial(
                        ham.build_TDM_aBJ,
                        TDM_matrix_element=partial(
                            me.TransitionDipole_odd_aBJ_noM, iH=iH, I=IM
                        ),
                    ),
                }
            if self.molecule == "CaOH":
                all_TDM = {
                    "40A000": partial(
                        ham.build_TDM_aBJ,
                        TDM_matrix_element=partial(
                            me.TransitionDipole_even_aBJ_noM, I=iH
                        ),
                    ),
                    "40B000": partial(
                        ham.build_TDM_aBJ,
                        TDM_matrix_element=partial(
                            me.TransitionDipole_even_aBJ_noM, I=iH
                        ),
                    ),
                }
        return all_TDM

    def collect_parity(self):
        iH = self.I_spins[-1]
        IM = self.I_spins[0]
        if self.molecule == "YbOH":
            all_parity = {
                "174X000": partial(
                    ham.build_operator, matrix_element=partial(me.Parity_L_bBJ, I=iH)
                ),
                "174X010": partial(
                    ham.build_operator, matrix_element=partial(me.Parity_l_bBJ, I=iH)
                ),
                "171X000": partial(
                    ham.build_operator,
                    matrix_element=partial(me.Parity_L_bBS, I=IM, iH=iH),
                ),
                "173X000": partial(
                    ham.build_operator,
                    matrix_element=partial(me.Parity_L_bBS, I=IM, iH=iH),
                ),
                "171X010": partial(
                    ham.build_operator,
                    matrix_element=partial(me.Parity_l_bBS, I=IM, iH=iH),
                ),
                "173X010": partial(
                    ham.build_operator,
                    matrix_element=partial(me.Parity_l_bBS, I=IM, iH=iH),
                ),
                "174A000": partial(
                    ham.build_operator, matrix_element=partial(me.Parity_L_aBJ, I=iH)
                ),
                "173A000": partial(
                    ham.build_operator,
                    matrix_element=partial(me.Parity_odd_L_aBJ, I=IM, iH=iH),
                ),
                "171A000": partial(
                    ham.build_operator,
                    matrix_element=partial(me.Parity_odd_L_aBJ, I=IM, iH=iH),
                ),
            }
        if self.molecule == "CaOH":
            all_parity = {
                "40X000": partial(
                    ham.build_operator, matrix_element=partial(me.Parity_L_bBJ, I=iH)
                ),
                "40X010": partial(
                    ham.build_operator, matrix_element=partial(me.Parity_l_bBJ, I=iH)
                ),
                "40B010": partial(
                    ham.build_operator, matrix_element=partial(me.Parity_l_bBJ, I=iH)
                ),
                "40A000": partial(
                    ham.build_operator, matrix_element=partial(me.Parity_L_aBJ, I=iH)
                ),
                "40B000": partial(
                    ham.build_operator, matrix_element=partial(me.Parity_L_bBJ, I=iH)
                ),
            }
        return all_parity

    def collect_alt_q(self, I_spins, P_values):
        if self.molecule == "YbOH":
            alt_q_builders = {
                "174X000": {
                    "aBJ": partial(
                        qn.q_numbers_even_aBJ, K_mag=0, I_list=I_spins, P_values=[1 / 2]
                    ),
                    "decoupled": partial(
                        qn.q_numbers_decoupled, K_mag=0, I_list=I_spins
                    ),
                },
                "174X010": {
                    "aBJ": partial(
                        qn.q_numbers_even_aBJ,
                        K_mag=1,
                        P_values=[1 / 2, 3 / 2],
                        I_list=I_spins,
                    ),
                    "decoupled": partial(
                        qn.q_numbers_decoupled, K_mag=1, I_list=I_spins
                    ),
                    "recouple_J": partial(
                        qn.q_numbers_decoupled_mJ, K_mag=1, I_list=I_spins
                    ),
                    "decouple_I": partial(
                        qn.q_numbers_decoupled_mJ, K_mag=1, I_list=I_spins
                    ),
                    "vibronic": partial(
                        qn.q_numbers_vibronic_even_aBJ,
                        l_mag=1,
                        L_mag=0,
                        I_list=I_spins,
                        P_values=[1 / 2, 3 / 2],
                    ),
                },
                "173X000": {
                    "aBJ": partial(
                        qn.q_numbers_odd_aBJ, K_mag=0, I_list=I_spins, P_values=[1 / 2]
                    ),
                    "bBJ": partial(qn.q_numbers_odd_bBJ, K_mag=0, I_list=I_spins),
                    "decoupled": partial(
                        qn.q_numbers_decoupled, K_mag=0, I_list=I_spins
                    ),
                },
                "173X010": {
                    "aBJ": partial(
                        qn.q_numbers_odd_aBJ,
                        K_mag=1,
                        P_values=[1 / 2, 3 / 2],
                        I_list=I_spins,
                    ),
                    "bBJ": partial(qn.q_numbers_odd_bBJ, K_mag=1, I_list=I_spins),
                    "decoupled": partial(
                        qn.q_numbers_decoupled, K_mag=1, I_list=I_spins
                    ),
                },
                "174A000": {
                    "bBS": partial(qn.q_numbers_bBS, K_mag=1, I_list=I_spins),
                    "bBJ": partial(qn.q_numbers_even_bBJ, K_mag=1, I_list=I_spins),
                    "decoupled": partial(
                        qn.q_numbers_decoupled, K_mag=1, I_list=I_spins
                    ),
                    # 'vibronic': partial(qn.q_numbers_vibronic_even_aBJ,l_mag=0,L_mag=1,
                    # I_list=I_spins,P_values=P_values),
                    # 'uA010': partial(qn.q_numbers_vibronic_even_aBJ,
                    # l_mag=1,L_mag=1,Omega_values=[1/2],
                    # K_values=[0],I_list=I_spins,P_values=P_values),
                    # 'kA010': partial(qn.q_numbers_vibronic_even_aBJ,
                    # l_mag=1,L_mag=1,Omega_values=[3/2],
                    # K_values=[0],I_list=I_spins,P_values=P_values),
                    # 'B010': partial(qn.q_numbers_vibronic_even_aBJ,l_mag=1,L_mag=0,
                    # I_list=I_spins,P_values=P_values),
                },
                "173A000": {
                    "bBS": partial(qn.q_numbers_bBS, K_mag=1, I_list=I_spins),
                    "bBJ": partial(qn.q_numbers_odd_bBJ, K_mag=1, I_list=I_spins),
                    "decoupled": partial(
                        qn.q_numbers_decoupled, K_mag=1, I_list=I_spins
                    ),
                },
                "171X000": {
                    "aBJ": partial(
                        qn.q_numbers_odd_aBJ, K_mag=0, I_list=I_spins, P_values=[1 / 2]
                    ),
                    "bBJ": partial(qn.q_numbers_odd_bBJ, K_mag=0, I_list=I_spins),
                    "decoupled": partial(
                        qn.q_numbers_decoupled, K_mag=0, I_list=I_spins
                    ),
                },
                "171X010": {
                    "aBJ": partial(
                        qn.q_numbers_odd_aBJ,
                        K_mag=1,
                        I_list=I_spins,
                        P_values=[1 / 2, 3 / 2],
                    ),
                    "bBJ": partial(qn.q_numbers_odd_bBJ, K_mag=1, I_list=I_spins),
                    "decoupled": partial(
                        qn.q_numbers_decoupled, K_mag=1, I_list=I_spins
                    ),
                },
                "171A000": {
                    "bBS": partial(qn.q_numbers_bBS, K_mag=1, I_list=I_spins),
                    "bBJ": partial(qn.q_numbers_odd_bBJ, K_mag=1, I_list=I_spins),
                    "decoupled": partial(
                        qn.q_numbers_decoupled, K_mag=1, I_list=I_spins
                    ),
                },
            }
        if self.molecule == "CaOH":
            alt_q_builders = {
                "40X000": {
                    "aBJ": partial(
                        qn.q_numbers_even_aBJ, K_mag=0, I_list=I_spins, P_values=[1 / 2]
                    ),
                    "decoupled": partial(
                        qn.q_numbers_decoupled, K_mag=0, I_list=I_spins
                    ),
                    "bBS": partial(qn.q_numbers_bBS, K_mag=0, I_list=I_spins),
                },
                "40X010": {
                    "aBJ": partial(
                        qn.q_numbers_even_aBJ,
                        K_mag=1,
                        P_values=[1 / 2, 3 / 2],
                        I_list=I_spins,
                    ),
                    "decoupled": partial(
                        qn.q_numbers_decoupled, K_mag=1, I_list=I_spins
                    ),
                    "bBS": partial(qn.q_numbers_bBS, K_mag=1, I_list=I_spins),
                },
                "40A000": {
                    "bBS": partial(qn.q_numbers_bBS, K_mag=1, I_list=I_spins),
                    "bBJ": partial(qn.q_numbers_even_bBJ, K_mag=1, I_list=I_spins),
                    "decoupled": partial(
                        qn.q_numbers_decoupled, K_mag=1, I_list=I_spins
                    ),
                    "recouple_J": partial(
                        qn.q_numbers_decoupled_mJ, K_mag=1, I_list=I_spins
                    ),
                },
                "40B000": {
                    "aBJ": partial(
                        qn.q_numbers_even_aBJ, K_mag=0, I_list=I_spins, P_values=[1 / 2]
                    ),
                    "decoupled": partial(
                        qn.q_numbers_decoupled, K_mag=0, I_list=I_spins
                    ),
                },
            }
        return alt_q_builders



import multiprocessing as mp
import numpy as np
from qutip import Qobj, mesolve, basis


def prep_sim(
    matrix_info,
    state_groups,
    qops,
    qd,
    s_X=1,
    s_Z=1,
    one_photon=0,
    e_decohere_rate=10 * 2 * np.pi,
    g_decohere_rate=0.05 * 2 * np.pi,
    decay_rate=1 / 0.02,
    vbr_scaling=None,
    time=np.linspace(0, 4, 5000),
    big_decohere=False,
    pol="XZ",
):

    H0, Hoff_Z, Hoff_X, BR0, BRp, BRm, rho0 = matrix_info
    e_states = state_groups[0]
    g_states = [g for group in state_groups[1:4] for g in group]

    Rabi_Z = decay_rate * np.sqrt(1 / 2 * s_Z) / 2
    Rabi_X = decay_rate * np.sqrt(1 / 2 * s_X) / 2

    if vbr_scaling is not None:
        Rabi_Z *= np.sqrt(vbr_scaling)
        Rabi_X *= np.sqrt(vbr_scaling)

    g_decohere = [np.sqrt(g_decohere_rate) * qops[i][i] for i in range(len(g_states))]
    e_decohere = [np.sqrt(e_decohere_rate) * qops[i][i] for i in range(len(e_states))]
    decay = [np.sqrt(decay_rate) * (Qobj(BR.T)) for BR in [BR0, BRp, BRm]]
    if not big_decohere:
        collapse = [*g_decohere, *e_decohere, *decay]
    else:
        collapse = [
            *g_decohere,
            *e_decohere,
            *decay,
            *[[qops[i][i] * np.sqrt(10), on_off2] for i in range(len(qops))],
        ]
    size = len(qops)
    Hd = np.zeros((size, size))
    for i in range(len(state_groups[0])):
        Hd[i, i] -= one_photon
    if pol == "XZ" or pol == "ZX":
        H = [
            Qobj(H0 + Hd) * 2 * np.pi,
            [Rabi_Z * Qobj(Hoff_Z + Hoff_Z.T), on_off],
            [Rabi_X * Qobj(Hoff_X), exp_onoff],
            [Rabi_X * Qobj(Hoff_X.T), expconj_onoff],
        ]
    else:
        H = [
            Qobj(H0 + Hd) * 2 * np.pi,
            [Rabi_Z * Qobj(Hoff_Z + Hoff_Z.T), on_off],
            [Rabi_X * Qobj(Hoff_Z), exp_onoff],
            [Rabi_X * Qobj(Hoff_Z.T), expconj_onoff],
        ]
    return H, rho0, time, collapse, state_groups


def mp_cpt_sim(
    two_photon_array, worker_func, H, rho0, time, collapse, state_groups, tf=3
):
    pool = mp.Pool(mp.cpu_count())
    results = pool.starmap(
        worker_func,
        [
            (two_photon, H, rho0, time, collapse, state_groups, tf)
            for two_photon in two_photon_array
        ],
    )
    pool.close()
    pool.join()
    result_array = np.array(results)
    return result_array


def prep_sim_1ph(
    one_photon_array,
    matrix_info,
    state_groups,
    qops,
    qd,
    s_X=1,
    s_Z=1,
    e_decohere_rate=10 * 2 * np.pi,
    g_decohere_rate=0.05 * 2 * np.pi,
    decay_rate=1 / 0.02,
    vbr_scaling=None,
    time=np.linspace(0, 4, 5000),
    big_decohere=False,
    pol="XZ",
):

    H0, Hoff_Z, Hoff_X, BR0, BRp, BRm, rho0 = matrix_info
    e_states = state_groups[0]
    g_states = [g for group in state_groups[1:4] for g in group]

    Rabi_Z = decay_rate * np.sqrt(1 / 2 * s_Z) / 2
    Rabi_X = decay_rate * np.sqrt(1 / 2 * s_X) / 2

    if vbr_scaling is not None:
        Rabi_Z *= np.sqrt(vbr_scaling)
        Rabi_X *= np.sqrt(vbr_scaling)

    g_decohere = [np.sqrt(g_decohere_rate) * qops[i][i] for i in range(len(g_states))]
    e_decohere = [np.sqrt(e_decohere_rate) * qops[i][i] for i in range(len(e_states))]
    decay = [np.sqrt(decay_rate) * (Qobj(BR.T)) for BR in [BR0, BRp, BRm]]
    if not big_decohere:
        collapse = [*g_decohere, *e_decohere, *decay]
    else:
        collapse = [
            *g_decohere,
            *e_decohere,
            *decay,
            *[[qops[i][i] * np.sqrt(10), on_off2] for i in range(len(qops))],
        ]

    size = len(qops)
    H_all = []
    for one_photon in one_photon_array:
        Hd = np.zeros((size, size))
        for i in range(len(state_groups[0])):
            Hd[i, i] -= one_photon
        if pol == "XZ" or pol == "ZX":
            H = [
                Qobj(H0 + Hd) * 2 * np.pi,
                [Rabi_Z * Qobj(Hoff_Z + Hoff_Z.T), on_off],
                [Rabi_X * Qobj(Hoff_X), exp_onoff],
                [Rabi_X * Qobj(Hoff_X.T), expconj_onoff],
            ]
        else:
            H = [
                Qobj(H0 + Hd) * 2 * np.pi,
                [Rabi_Z * Qobj(Hoff_Z + Hoff_Z.T), on_off],
                [Rabi_X * Qobj(Hoff_Z), exp_onoff],
                [Rabi_X * Qobj(Hoff_Z.T), expconj_onoff],
            ]
        H_all.append(H)
    return H_all, rho0, time, collapse, state_groups


def mp_cpt_sim_1ph(
    one_photon_array,
    two_photon,
    worker_func,
    H,
    rho0,
    time,
    collapse,
    state_groups,
    tf=3,
):
    pool = mp.Pool(mp.cpu_count())
    results = pool.starmap(
        worker_func,
        [(two_photon, H_, rho0, time, collapse, state_groups, tf) for H_ in H],
    )
    pool.close()
    pool.join()
    result_array = np.array(results)
    return result_array


def cpt_sim(two_photon, H, rho0, time, collapse, state_groups, tf):
    two_photon = two_photon * 2 * np.pi
    result = mesolve(
        H,
        rho0,
        time,
        collapse,
        [sum(states) for states in state_groups],
        args={"on": 0, "off": tf, "on2": 3.2, "off2": 4, "w": two_photon},
    )
    return result.expect


def prep_sim_1d_weighted(
    matrix_info,
    state_groups,
    qops,
    qd,
    s_X=1,
    s_Z=1,
    one_photon=0,
    e_decohere_rate=10 * 2 * np.pi,
    g_decohere_rate=0.05 * 2 * np.pi,
    decay_rate=1 / 0.02,
    vbr_scaling=None,
    time=np.linspace(0, 4, 5000),
    big_decohere=False,
):

    H0, Hoff_Z, Hoff_X, BR0, BRp, BRm, rho0 = matrix_info
    e_states = state_groups[0]
    g_states = [g for group in state_groups[1:4] for g in group]

    Rabi_Z = decay_rate * np.sqrt(1 / 2 * s_Z) / 2
    Rabi_X = decay_rate * np.sqrt(1 / 2 * s_X) / 2

    if vbr_scaling is not None:
        Rabi_Z *= np.sqrt(vbr_scaling)
        Rabi_X *= np.sqrt(vbr_scaling)

    g_decohere = [np.sqrt(g_decohere_rate) * qops[i][i] for i in range(len(g_states))]
    e_decohere = [np.sqrt(e_decohere_rate) * qops[i][i] for i in range(len(e_states))]
    decay = [np.sqrt(decay_rate) * (Qobj(BR.T)) for BR in [BR0, BRp, BRm]]
    if not big_decohere:
        collapse = [*g_decohere, *e_decohere, *decay]
    else:
        collapse = [
            *g_decohere,
            *e_decohere,
            *decay,
            *[[qops[i][i] * np.sqrt(10), on_off2] for i in range(len(qops))],
        ]

    size = len(qops)
    Hd = np.zeros((size, size))
    for i in range(len(state_groups[0])):
        Hd[i, i] -= one_photon

    H = Qobj(H0 + Hd) * 2 * np.pi
    # Hall = []
    Hinfo = [H, Hoff_Z, Hoff_X, Rabi_X, Rabi_Z]
    # for s_ in s_array:
    #     Hs = [H,[decay_rate*np.sqrt(1/2*s_)/2*Qobj(Hoff_Z + Hoff_Z.T),on_off],
    # [decay_rate*np.sqrt(1/2*s_)/2*Qobj(Hoff_X),exp_onoff],[decay_rate*np.sqrt(1/2*s_)
    # /2*Qobj(Hoff_X.T),expconj_onoff]]
    #     Hall.append(Hs)

    return Hinfo, rho0, time, collapse, state_groups


def mp_cpt_sim_1d_weighted(
    two_photon_array,
    s_weight_array,
    worker_func,
    Hinfo,
    rho0,
    time,
    collapse,
    state_groups,
    tf=3,
    pol="XZ",
):
    pool = mp.Pool(mp.cpu_count())
    results = pool.starmap(
        worker_func,
        [
            (two_photon, s_weight, Hinfo, rho0, time, collapse, state_groups, tf, pol)
            for two_photon, s_weight in zip(two_photon_array, s_weight_array)
        ],
    )
    pool.close()
    pool.join()
    result_array = np.array(results)
    return result_array


def cpt_sim_1d_weighted(
    two_photon, s_weight, Hinfo, rho0, time, collapse, state_groups, tf, pol
):
    two_photon = two_photon * 2 * np.pi

    H0, Hoff_Z, Hoff_X, Rabi_X, Rabi_Z = Hinfo
    if pol == "XZ":
        H = [
            H0,
            [Rabi_Z * Qobj(Hoff_Z + Hoff_Z.T), on_off],
            [Rabi_X * np.sqrt(s_weight) * Qobj(Hoff_X), exp_onoff],
            [Rabi_X * np.sqrt(s_weight) * Qobj(Hoff_X.T), expconj_onoff],
        ]
    else:
        H = [
            H0,
            [Rabi_Z * Qobj(Hoff_Z + Hoff_Z.T), on_off],
            [Rabi_X * np.sqrt(s_weight) * Qobj(Hoff_Z), exp_onoff],
            [Rabi_X * np.sqrt(s_weight) * Qobj(Hoff_Z.T), expconj_onoff],
        ]
    result = mesolve(
        H,
        rho0,
        time,
        collapse,
        [sum(states) for states in state_groups],
        args={"on": 0, "off": tf, "w": two_photon},
    )
    return result.expect


def prep_sim2d(
    s_array,
    matrix_info,
    state_groups,
    qops,
    qd,
    one_photon=0,
    e_decohere_rate=10 * 2 * np.pi,
    g_decohere_rate=0.05 * 2 * np.pi,
    decay_rate=1 / 0.02,
    vbr_scaling=None,
    time=np.linspace(0, 4, 5000),
    big_decohere=False,
):

    H0, Hoff_Z, Hoff_X, BR0, BRp, BRm, rho0 = matrix_info
    e_states = state_groups[0]
    g_states = [g for group in state_groups[1:4] for g in group]
    if len(s_array) == 2:
        s_X, s_Z = s_array
    elif len(s_array) == 1:
        s_X = s_array[0]
        s_Z = s_array[0]
    else:
        s_X, s_Y, s_Z = s_array
    Rabi_Z = decay_rate * np.sqrt(1 / 2 * s_Z) / 2
    Rabi_X = decay_rate * np.sqrt(1 / 2 * s_X) / 2

    if vbr_scaling is not None:
        Rabi_Z *= np.sqrt(vbr_scaling)
        Rabi_X *= np.sqrt(vbr_scaling)

    g_decohere = [np.sqrt(g_decohere_rate) * qops[i][i] for i in range(len(g_states))]
    e_decohere = [np.sqrt(e_decohere_rate) * qops[i][i] for i in range(len(e_states))]
    decay = [np.sqrt(decay_rate) * (Qobj(BR.T)) for BR in [BR0, BRp, BRm]]
    if not big_decohere:
        collapse = [*g_decohere, *e_decohere, *decay]
    else:
        collapse = [
            *g_decohere,
            *e_decohere,
            *decay,
            *[[qops[i][i] * np.sqrt(10), on_off2] for i in range(len(qops))],
        ]

    size = len(qops)
    Hd = np.zeros((size, size))
    for i in range(len(state_groups[0])):
        Hd[i, i] -= one_photon

    H = Qobj(H0 + Hd) * 2 * np.pi
    # Hall = []
    Hinfo = [H, Hoff_Z, Hoff_X, decay_rate]
    # for s_ in s_array:
    #     Hs = [H,[decay_rate*np.sqrt(1/2*s_)/2*Qobj(Hoff_Z + Hoff_Z.T),on_off],
    # [decay_rate*np.sqrt(1/2*s_)/2*Qobj(Hoff_X),exp_onoff],[decay_rate*np.sqrt(1/2*s_)/
    # 2*Qobj(Hoff_X.T),expconj_onoff]]
    #     Hall.append(Hs)

    return Hinfo, rho0, time, collapse, state_groups


def mp_cpt_sim_2d(
    two_photon_array,
    s_array,
    worker_func,
    Hinfo,
    rho0,
    time,
    collapse,
    state_groups,
    tf=3,
    pol="XZ",
):
    mesh = np.array(np.meshgrid(two_photon_array, s_array, indexing="ij"))
    pairs = np.stack((mesh[0].ravel(), mesh[1].ravel()), axis=1)
    pool = mp.Pool(mp.cpu_count())
    results = pool.starmap(
        worker_func,
        [
            (two_photon, s_, Hinfo, rho0, time, collapse, state_groups, tf, pol)
            for two_photon, s_ in pairs
        ],
    )
    pool.close()
    pool.join()
    result_array = np.array(results)
    return result_array


# def cpt_sim_2d(two_photon, s ,Hinfo,rho0,time,collapse,state_groups,tf,pol):
#     two_photon = two_photon*2*np.pi
#     H0, Hoff_Z, Hoff_X, decay_rate = Hinfo
#     if pol=='XZ':
#         H = [H0,[decay_rate*np.sqrt(1/2*s)/2*Qobj(Hoff_Z + Hoff_Z.T),on_off],
# [decay_rate*np.sqrt(1/2*s)/2*Qobj(Hoff_X),exp_onoff],[decay_rate*np.sqrt(1/2*s)/2*
# Qobj(Hoff_X.T),expconj_onoff]]
#     else:
#         H = [H0,[decay_rate*np.sqrt(1/2*s)/2*Qobj(Hoff_Z + Hoff_Z.T),on_off],[decay_rate*
# np.sqrt(1/2*s)/2*Qobj(Hoff_Z),exp_onoff],[decay_rate*np.sqrt(1/2*s)/2*Qobj(Hoff_Z.T),expconj_onoff]]
#     result = mesolve(H,rho0,time,collapse,[sum(states) for states in state_groups],args={'on':0,
# 'off':tf,'on2':3.2,'off2':4,'w':two_photon})
#     return result.expect


def cpt_sim_2d(two_photon, s, Hinfo, rho0, time, collapse, state_groups, tf, pol):
    two_photon = two_photon * 2 * np.pi
    H0, Hoff_Z, Hoff_X, decay_rate = Hinfo
    if pol == "XZ":
        H = [
            H0,
            [decay_rate * np.sqrt(1 / 2 * s) / 2 * Qobj(Hoff_Z + Hoff_Z.T), on_off],
            [decay_rate * np.sqrt(1 / 2 * s) / 2 * Qobj(Hoff_X), exp_onoff],
            [decay_rate * np.sqrt(1 / 2 * s) / 2 * Qobj(Hoff_X.T), expconj_onoff],
        ]
    else:
        H = [
            H0,
            [decay_rate * np.sqrt(1 / 2 * s) / 2 * Qobj(Hoff_Z + Hoff_Z.T), on_off],
            [decay_rate * np.sqrt(1 / 2 * s) / 2 * Qobj(Hoff_Z), exp_onoff],
            [decay_rate * np.sqrt(1 / 2 * s) / 2 * Qobj(Hoff_Z.T), expconj_onoff],
        ]
    result = mesolve(
        H,
        rho0,
        time,
        collapse,
        [sum(states) for states in state_groups],
        args={"on": 0, "off": tf, "on2": 3.2, "off2": 4, "w": two_photon},
    )
    return result.expect


# generate n level basis
def q_basis(n):
    q_basis = []
    for i in range(n):
        q_basis.append(basis(n, i))
    return q_basis


def q_dict(q_basis, q_str):
    q_dict = {qs: qb for qs, qb in zip(q_str, q_basis)}
    return q_dict


def q_ops(q_basis):
    n = len(q_basis)
    q_ops = np.zeros((n, n)).tolist()
    for i in range(n):
        for j in range(n):
            q_ops[i][j] = q_basis[i] * q_basis[j].dag()
    return q_ops


def exp_osc(t, args):
    w = args["w"]
    return np.exp(1j * w * t)


def exp_conj(t, args):
    w = args["w"]
    return np.exp(-1j * w * t)


def linear_ramp(t, args):
    m = args["slope"]
    b = args["start"]
    stop = args["stop"]
    y = m * t + b
    if y > stop:
        return 0
    else:
        return y


def on_off(t, args):
    on = args["on"]
    off = args["off"]
    if t > on and t < off:
        return 1
    else:
        return 0


def on_off2(t, args):
    on = args["on2"]
    off = args["off2"]
    if t > on and t < off:
        return 1
    else:
        return 0


def exp_onoff(t, args):
    w = args["w"]
    on = args["on"]
    off = args["off"]
    exp = np.exp(1j * w * t)
    if t > on and t < off:
        return 1 * exp
    else:
        return 0


def expconj_onoff(t, args):
    w = args["w"]
    on = args["on"]
    off = args["off"]
    exp = np.exp(-1j * w * t)
    if t > on and t < off:
        return 1 * exp
    else:
        return 0



# Here we will test the basic functionalities of the package
import unittest
from molecular_structure import MoleculeLevels


class TestCalculations(unittest.TestCase):

    def test_sum(self):
        print("Testing MoleculeLevels calculation")
        calculation = MoleculeLevels.initialize_state(
            "YbOH",
            "174",
            "X000",
            N_range=[0, 1, 2, 3],
            M_values="all",
            I=[0, 1 / 2],
            P_values=[1 / 2],
            S=1 / 2,
            round=6,
        )
        self.assertEqual(calculation is not None, True)


if __name__ == "__main__":
    unittest.main()
